/************************************************/
/*						*/
/*      Copyright:				*/
/*		 Jean-Marc Pigeon <jmp@safe.ca>	*/
/*	Distributed under the Gnu Public	*/
/*	License, see the LICENCE file in this	*/
/*	package.				*/
/*						*/
/*	Define low level UTiLity routine to do a*/
/*	flexible memory garbage collector.	*/
/*						*/
/************************************************/
#ifndef UTLGAC
#define UTLGAC
#include	<stdlib.h>
#include	<stdarg.h>


/*to activate or stop memory monitoring		*/
typedef enum {gear_down,gear_up}GEAR;

typedef	void GACTYP;

/*administrative routine			*/
extern const char *gac_getversion();
extern void gac_setcurbase(int *debug,int *verbose,char *path);
extern void gac_setlogdir(const char *path);


/*monitoring  routine				*/
/*to open a memory monitoring area.		*/
extern GACTYP *gac_open(int tgtdbg,const char *apname,const char *extension);

/*to close a memory monitoring area.		*/
extern int gac_close(GACTYP *gac,long *remaining);

/*to free all memory remaining within an  area	*/
/*and clode the memory monitoring area		*/
/*return 0 if everything fine, -1 if meemory	*/
/*still found assigned				*/
extern int gac_trash(int tgtdbg,GACTYP *gac,long *remaining);

/*to control which garbage collector will check	*/
/*memory, if gear up is done on the last opened	*/
/*GC, memory is not check at all, same if gear	*/
/*down reach a down level bellow first opened	*/
/*GC, memory is not supervised.			*/
extern int gac_gear(GEAR updown,long *remaining);

/*procedure to write a message on the CURRENT	*/
/*garbage collector report file. Return the	*/
/*number of character written in the report	*/
/*file.						*/
extern int gac_report(int tgtdbg,char *fmt,...);

/*routine we do NOT want to be glibc one 	*/
/*Primary memory function			*/
extern void *malloc(size_t size);
extern void *realloc(void *ptr,size_t size);
extern void free(void *ptr);
extern void *memaligne(size_t alignement,size_t size);
extern void *calloc(size_t nmemb, size_t size);

/*routine we do NOT want to be glibc one 	*/
/*Secondary memory function			*/
extern int putenv(char *valeur);
extern int setenv(const char *name,const char *value,int overwrite);
extern int unsetenv(const char *name);
extern int vasprintf(char **strp, const char *fmt,va_list args);
extern int asprintf(char **strp, const char *fmt, ...);
extern char *strdup(const char *str);
extern char *strndup(const char *str,size_t n);
#endif
