/************************************************/
/*						*/
/*      Copyright:				*/
/*		 Jean-Marc Pigeon <jmp@safe.ca>	*/
/*	Distributed under the Gnu Public	*/
/*	License, see the LICENCE file in this	*/
/*	package.				*/
/*						*/
/*	Implement a garbage collector for memory*/
/*	assigned.				*/
/*						*/
/************************************************/
#include	<stdlib.h>
#include	<signal.h>
#include	<stdio.h>
#include	<dlfcn.h>
#include	<errno.h>
#include	<string.h>
#include	<malloc.h>
#include	<string.h>
#include	<unistd.h>
#include	<time.h>
#include	<search.h>
#include	<execinfo.h>
#include	<sys/time.h>
#include	<sys/file.h>
#include	<utlgac.h>

/************************************************/
/*	Including version number		*/
/************************************************/
#include	"curtag.tag"

/********************************************************/
#define MXTB	20	/*max trace deep		*/
#define	MXD	400	/*maximun path size		*/
#define	OFFTB	3	/*trace back offset		*/

/*defining boolean value				*/
typedef enum {false,true} BOOL;

/*you should alway provide an application name 		*/
/*to function gac_open					*/
#define	APPNAME	"Application_name"

#define	LOGDIR	"/var/tmp"

typedef	struct timeval	TIMEVAL;

typedef struct 	{	/*leak detection structure	*/
	void *ptr;	/*allocated memory		*/
	size_t size;	/*memory size			*/
	int upgac;	/*defined on previous gac	*/
	TIMEVAL date;	/*allocation date		*/
	void *tb[MXTB]; /*back trace			*/
	}LEAKTYP;

typedef struct gacmem GACMEM;

struct	gacmem {
	int ready;	/*GC ready to work		*/
	long memstart;	/*mallinfo memory used at start*/
	u_long tomemo;	/*number of memory given	*/
	u_long tosize;	/*total memory size		*/
	int prvhookset;	/*previous hookset		*/
	int tgtdbg;	/*Garbage Collector debug level	*/
	char *appname;	/*application using GC		*/
	char *extension;/*app/extension using GC	*/
	void *memory;	/*given memory tree node	*/
	GACMEM *next;	/*next garbage collector	*/
	GACMEM *prev;	/*previous garbage collector	*/
	};

/*garbage collector (GC) definition			*/
static	GACMEM	*cgac;	/*current running GC		*/
static	GACMEM	*gaclst=(GACMEM *)0;

/*To trace in Garbage Collector memory			*/
static u_long collected;/*# memory found open		*/
static LEAKTYP **memsort;

static char cpath[MXD];	/*Application curpath		*/
static char cbase[MXD];	/*Application base		*/
static char logdir[MXD];/*log directory			*/

/*code area definition					*/
static void *codefirst;	/*code first address		*/
static void *codelast;	/*code last address		*/
static int inhook=false;/*out of hookmode		*/

static int lvbose=0;	/*local verbose (always=false)	*/
static int ldebug=0;	/*local debug (always=0)	*/
static int *dbg=&ldebug;/*application debug level	*/
static int *vbo=&lvbose;/*application verbose level	*/

/*

*/
/************************************************/
/*						*/
/*	Procedure to close a report file	*/
/*						*/
/************************************************/
static int closerptfile(FILE *report)

{
int status;

status=0;
if (report!=stderr)
  status=fclose(report);
return status;
}
/*

*/
/************************************************/
/*						*/
/*	Procedure to open the report file	*/
/*	if can't use stderr			*/
/*						*/
/************************************************/
static FILE *openrptfile(GACMEM *gac)

{
FILE *report;

report=stderr;
if ((*vbo==false)&&(strlen(logdir)>0)&&(gac!=(GACMEM *)0)) {
  char *logname;
  char *sep;
  char *locapp;
  char *locext;

  sep="";
  locapp=APPNAME;
  locext="";
  if ((gac->extension!=(char *)0)&&(strlen(gac->extension)>0)) {
    sep="-";
    locext=gac->extension;
    }
  if ((gac->appname!=(char *)0)&&(strlen(gac->appname)>0)) 
    locapp=gac->appname;
  (void) asprintf(&logname,"%s/%s-%s%s%s-%08d.jrl",logdir,"memoryleak",locapp,sep,locext,getpid()); 
  if ((report=fopen(logname,"a+"))==(FILE *)0) {
    (void) fprintf(stderr,"%s, Unable to open file <%s>, (errno=<%s>\n",
			  "utlgac.c:openrptfile,",logname,strerror(errno));
    (void) sleep(2);	/*to avoid avalanche	*/
    (void) kill(getpid(),SIGSEGV);/*core dump	*/
    (void) exit(-1);	/*never reached		*/
    }
  (void) free(logname);
  }
return report;
}
/*

*/
/************************************************/
/*						*/
/*	Procedure to convert back trace address	*/
/*	to sources line information.		*/
/*						*/
/************************************************/
static void disline(LEAKTYP *leak,const char *path,FILE *fichier,char *extension)

{
#define ADDR2LINE	"addr2line"

int i;

(void) fprintf(fichier,"Trace back calling procedures (module=<%s>)\n",extension);
for (i=0;leak->tb[i]!=(void *)0;i++) {
  char *command;
  Dl_info info;

  command=(char *)0;
  if (dladdr(leak->tb[i],&info)==0) {
    (void) fprintf(fichier,"Unable to find address information for address '%p' (error=<%s> bug?)\n",
			   leak->tb[i],dlerror());
    continue;
    }
  if ((codefirst<=leak->tb[i])&&(leak->tb[i]<=codelast)) {
    FILE *canal;

    canal=(FILE *)0;
    (void) asprintf(&command,"%s -C -i -e %s %p",ADDR2LINE,cpath,leak->tb[i]);
    if ((canal=popen(command,"r"))!=(FILE *)0) {
      char line[200];

      while (fgets(line,sizeof(line)-1,canal)!=(char *)0) {
	if (fichier!=(FILE *)0)
          (void) fprintf(fichier,"%p:\t%s",leak->tb[i],line);
        }
      (void) fclose(canal);
      }
    }
  else
    (void) fprintf(fichier,"%p:\t%s\n",(void *)((char *)leak->tb[i]-(char *)info.dli_fbase),info.dli_fname);
  (void) free(command);
  }
}
/*

*/
/************************************************/
/*						*/
/*	Procedure to assigne a Backtrace stack	*/
/*	(to trace back leak) to memory structure*/
/*						*/
/************************************************/
static void setbtrace(LEAKTYP *leak,int step)

{
void *zone[MXTB];
int num;
int i,j;
struct timezone tz;

(void) memset(zone,'\000',sizeof(zone));
(void) memset(leak->tb,'\000',sizeof(leak->tb));
num=backtrace(zone,MXTB-1);
for (j=0,i=step+1;i<num;i++,j++) {
  leak->tb[j]=zone[i];
  }
(void) gettimeofday(&(leak->date),&tz);
}
/*

*/
/************************************************/
/*						*/
/*	Procedure to report processus trace back*/
/*						*/
/************************************************/
static void traceback(int tgtdbg,int step,char *extension,FILE *report)

{
if (*dbg>=tgtdbg) {
  LEAKTYP leak;

  (void) setbtrace(&leak,step+3);
  (void) disline(&leak,cpath,report,extension);
  }
}
/*

*/
/************************************************/
/*						*/
/*	Procedure to report a trouble.		*/
/*						*/
/************************************************/
static void trblreport(GACMEM *gac,char *str)

{
if ((gac!=(GACMEM *)0)&&(*dbg>=gac->tgtdbg)) {
  FILE *report;

  report=openrptfile(gac);
  (void) flock(fileno(report),LOCK_EX);
  (void) fprintf(report,"%s\n",str);
  (void) traceback(gac->tgtdbg,OFFTB-1,gac->extension,report);
  (void) flock(fileno(report),LOCK_UN);
  (void) closerptfile(report);
  }
}
/*

*/
/************************************************/
/*						*/
/*	Procedure to remove a GC definition from*/
/*	The list.				*/
/*	Return -1 if trouble			*/
/*						*/
/************************************************/
static int remgac(GACMEM **lsts,GACMEM *gac)

{
int status;
int phase;
int proceed;

status=-1;
phase=0;
proceed=true;
while (proceed==true) {
  switch (phase) {
    case 0	:	/*scanning list? 	*/
      while (((*lsts)!=(GACMEM *)0)&&((*lsts!=gac)))
	lsts=&((*lsts)->next);
      if (*lsts==(GACMEM *)0)
	phase=999;	/*gac not found!	*/
      break;
    case 1	:	/*extract gac		*/
      (*lsts)=gac->next;
      if (*lsts!=(GACMEM *)0) 
	(*lsts)->prev=gac->prev;
      status=0;
      break;
    default	:	/*SAFE Guard		*/
      proceed=false;
      break;
    }
  phase++;
  }
return status;
}
/*

*/
/************************************************/
/*						*/
/*	Procedure to add a GACMEM structure to	*/
/*	list.					*/
/*						*/
/************************************************/
static void addinggac(GACMEM **lsts,GACMEM *gac)

{
if (*lsts==(GACMEM *)0)
  *lsts=gac;
else {
  register GACMEM *ptr;

  ptr=*lsts;
  while (ptr->next!=(GACMEM *)0) 
    ptr=ptr->next;
  ptr->next=gac;
  gac->prev=ptr;
  }
}
/*

*/
/************************************************/
/*						*/
/*	Procedure to compare to "leaking" memory*/
/*	ptr, used by search routine		*/
/*						*/
/************************************************/
static int cmpptr(const void *v1,const void *v2)

{
register u_long p1;
register u_long p2;

if (v1==(void *)0) {
  if (v2==(void *)0)
    return 0;
  else
    return -1;
  }
if (v2==(void *)0)
  return 1;
p1=(u_long)(((LEAKTYP *)v1)->ptr);
p2=(u_long)(((LEAKTYP *)v2)->ptr);
if (p1>p2)
  return 1;
if (p1<p2)
  return -1;
return 0;
}
/*

*/
/************************************************/
/*						*/
/*	Procedure to compare to "leaking" memory*/
/*	information, from the backtrace stand	*/
/*	point.					*/
/*						*/
/************************************************/
static int sametrace(const LEAKTYP *l1,const LEAKTYP *l2)

{
int inx;
long diff;

inx=0;
diff=0;
while ((inx<MXTB)&&(diff==(long)0)) {
  diff=(long)(l1->tb[inx])-(long)(l2->tb[inx]);
  inx++;
  }
return diff;
}
/*

*/
/************************************************/
/*						*/
/*	Procedure to compare to "leaking" memory*/
/*	information				*/
/*						*/
/************************************************/
static int cmpleak(const void *p1,const void *p2)

{
int diff;
LEAKTYP *l1;
LEAKTYP *l2;

l1=((LEAKTYP **)p1)[0];
l2=((LEAKTYP **)p2)[0];
if ((diff=sametrace(l1,l2))==0)
  diff=l1->date.tv_sec-l2->date.tv_sec;
  if (diff==0)
    diff=l1->date.tv_usec-l2->date.tv_usec;
return diff;
}
/*

*/
/************************************************/
/*						*/
/*	Procedure to count collected "leaking"	*/
/*	memory.					*/
/*						*/
/************************************************/
static void collectleakptr(const void *v1,const VISIT which,const int depth)

{
switch (which) {
  case preorder:	/*no remote dump	*/
  case endorder:
    break;
  case postorder:	/*remote dump		*/
  case leaf:
    memsort[collected]=*((LEAKTYP **)v1);
    collected++;
    break;
  }
}
/*

*/
/************************************************/
/*						*/
/*	procedure to get a "leak proof" memory	*/
/*	email to a shell script			*/
/*						*/
/************************************************/
static void setmemory(GACMEM *gac,void *ptr,size_t size)

{
register LEAKTYP *leak;
register void **leakptr;

leak=(LEAKTYP *)calloc(1,sizeof(LEAKTYP));
leak->size=size;
leak->ptr=ptr;
(void) setbtrace(leak,OFFTB);
gac->tomemo++;
gac->tosize+=size;
leakptr=tsearch((void *)leak,&(gac->memory),cmpptr);
if (*leakptr!=leak) {
  char *strloc;

  (void) free(leak);
  leak=*leakptr;
  gac->tomemo--;
  gac->tosize-=leak->size;
  leak->size=size;
  leak->ptr=ptr;
  (void) setbtrace(leak,OFFTB);
  (void) asprintf(&strloc,"process[%05d], %s-%s memory 'already set' ptr for ptr '%p'",
			  getpid(),gac->appname,gac->extension,ptr);
  (void) trblreport(gac,strloc);
  (void) free(strloc);
  }
}
/*

*/
/************************************************/
/*						*/
/*	Procedure to reallocate memory from pole*/
/*	while checking (if needed) for memory	*/
/*	leakage.				*/
/*						*/
/************************************************/
static void updatememory(GACMEM *gac,void *newptr,void *prvptr,size_t size)

{
if (prvptr==(void *)0)
  (void) setmemory(cgac,newptr,size);
else {
  LEAKTYP **leakpt;
  LEAKTYP leak;

  leak.ptr=prvptr;
  if ((leakpt=(LEAKTYP **)tfind(&leak,&(gac->memory),cmpptr))!=(LEAKTYP **)0) {
    LEAKTYP *lk;

    lk=*leakpt;
    (void) tdelete(*leakpt,&(gac->memory),cmpptr);
    gac->tosize-=lk->size;
    lk->ptr=newptr;
    lk->size=size;
    gac->tosize+=lk->size;
    (void) setbtrace(lk,OFFTB);
    (void) tsearch(lk,&(gac->memory),cmpptr);
    }
  else {
    char strloc[2000];

    (void) snprintf(strloc,sizeof(strloc),"process[%05d], %s-%s Unable to find 'realloc memory' previous ptr='%p'",
			  		   getpid(),gac->appname,gac->extension,prvptr);
    (void) trblreport(gac,strloc);
    (void) setmemory(cgac,newptr,size);
    }
  }
}
/*

*/
/************************************************/
/*						*/
/*	Procedure to free memory from pool	*/
/*	while checking (if needed) for memory	*/
/*	leakage.				*/
/*						*/
/************************************************/
void freememory(GACMEM *gac,void *ptr)

{
if (ptr!=(void *)0) {
  LEAKTYP **leakpt;
  LEAKTYP leak;

  leak.ptr=ptr;
  if ((leakpt=(LEAKTYP **)tfind(&leak,&(gac->memory),cmpptr))!=(LEAKTYP **)0) {
    LEAKTYP *lk;

    lk=*leakpt;
    (void) tdelete(*leakpt,&(gac->memory),cmpptr);
    gac->tomemo--;
    gac->tosize-=lk->size;
    (void) free(lk);
    }
  else {
    char *strloc;

    (void) asprintf(&strloc,"process[%05d], %s-%s, Unable to find 'to be freed memory' ptr='%p'",
			    getpid(),gac->appname,gac->extension,ptr);
    (void) trblreport(gac,strloc);
    (void) free(strloc);
    }
  }
}
/*

*/
/************************************************/
/*						*/
/*	Procedure to get program code first	*/
/*	and last address,			*/
/*						*/
/************************************************/
static void setcodelim()

{
extern void __executable_start;
extern void __etext;

codefirst=&__executable_start;
codelast=&__etext;;
}
/*

*/
/************************************************/
/*						*/
/*	Procedure to switch hook mode, getting	*/
/*	previous function caller.		*/
/*	to decide if call to garbage collector	*/
/*	accounting is needed.			*/
/*						*/
/************************************************/
static int switchhook(int *prvhook,const void *caller)

{
int needgc;

*prvhook=inhook;
needgc=inhook;
inhook=false;
if (needgc) {
  if ((caller>codelast)||(caller<codefirst))
    needgc=false;
  }
return needgc;
}
/*

*/
/************************************************/
/*						*/
/*	Procedure to allocate memory space	*/
/*	and log it in the garbage collector.	*/
/*						*/
/************************************************/
static void *gacmalloc(const void *caller,size_t size)

{
static void *(*r_malloc)(size_t)=(void *(*)(size_t))0;

register       void *result;
	       int hookset;

if (r_malloc==(void *(*)(size_t))0) 
  *((void **)&r_malloc)=dlsym(RTLD_NEXT,"malloc");
result=r_malloc(size);
if (switchhook(&hookset,caller)==true)
  (void) setmemory(cgac,result,size);
inhook=hookset;
return result;
}
/*

*/
/************************************************/
/*						*/
/*	Procedure to reallocate memory space	*/
/*	and log it in the garbage collector.	*/
/*						*/
/************************************************/
static void *gacrealloc(const void *caller,void *ptr,size_t size)

{
static void *(*r_realloc)(void *,size_t)=(void *(*)(void *,size_t))0;

register       void *result;
	       int hookset;

if (r_realloc==(void *(*)(void*,size_t))0) 
  *((void **)&r_realloc)=dlsym(RTLD_NEXT,"realloc");
result=r_realloc(ptr,size);
if (switchhook(&hookset,caller)==true) 
  (void) updatememory(cgac,result,ptr,size);
inhook=hookset;
return result;
}
/*

*/
/************************************************/
/*						*/
/*	Procedure to free memory space		*/
/*	and log it out from the garbage 	*/
/*	collector.				*/
/*						*/
/************************************************/
static void gacfree(const void *caller,void *ptr)

{
static void *(*r_free)(void *)=(void *(*)(void *))0;

int hookset;

if (r_free==(void *(*)(void *))0) 
  *((void **)&r_free)=dlsym(RTLD_NEXT,"free");
if (switchhook(&hookset,caller)==true)
  (void) freememory(cgac,ptr);
(void) r_free(ptr);
inhook=hookset;
}
/*

*/
/************************************************/
/*						*/
/*	Procedure to allocate memory properly	*/
/*	aligned onf according boundery multiple	*/
/*						*/
/************************************************/
static void *gacmemaligne(const void *caller,size_t alignement,size_t size)

{
static void *(*r_memalign)(size_t,size_t)=(void *(*)(size_t,size_t))0;

register void *result;
int hookset;

if (r_memalign==(void *(*)(size_t,size_t))0) 
  *((void **)&r_memalign)=dlsym(RTLD_NEXT,"memalign");
result=r_memalign(alignement,size);
if (switchhook(&hookset,caller)==true)
  (void) setmemory(cgac,result,size);
inhook=hookset;
return result;
}
/*

*/
/************************************************/
/*						*/
/*	Procedure to get a memory area fill with*/
/*	zero.					*/
/*						*/
/************************************************/
static void *gaccalloc(const void *caller,size_t nmemb,size_t size)

{
static void *(*r_calloc)(size_t,size_t)=(void *(*)(size_t,size_t))0;

register void *result;
int hookset;

if (r_calloc==(void *(*)(size_t,size_t))0) {
  /* Jakub Jelinek  7 Aug 200
   * dlsym may call calloc, in which case we'd end up in endless
   * recursion. No threads should be running at this point, since
   * pthread_initialize_manager calls malloc. If calloc returns
   * NULL in _dlerror_run, a static buffer will be used which is
   * fine, since no threads are running yet.
   */
  static int breakloop;

  if (breakloop)
    return (void *)0;
  breakloop=1;
  *((void **)&r_calloc)=dlsym(RTLD_NEXT,"calloc");
  breakloop=0;
  }
result=r_calloc(nmemb,size);
if (switchhook(&hookset,caller)==true)
  (void) setmemory(cgac,result,nmemb*size);
inhook=hookset;
return result;
}
/*

*/
/************************************************/
/*						*/
/*	Procedure to remove a variable from the	*/
/*	environnement list			*/
/*						*/
/************************************************/
static int gacunsetenv(const void *caller,const char *name)

{
int status;

status=-1;
errno=EINVAL;
if ((name!=(char *)0)&&(strchr(name,'=')==(char *)0)) {
  if (environ!=(char **)0) {
    int i;

    status=0;	/*success even if not existing	*/
    for (i=0;environ[i]!=(char *)0;i++) {
      char *ptr;
  
      if ((ptr=strstr(environ[i],name))!=environ[i])
        continue;
      if (*(environ[i]+strlen(name))=='=') {
        (void) gacfree(caller,ptr);
        do {
          environ[i]=environ[i+1];
	  i++;
	  }
        while (environ[i]!=(char *)0);
        break;
        }
      }
    }
  }
return status;
}
/*

*/
/************************************************/
/*						*/
/*	Procedure to install variable within the*/
/*	environnement				*/
/*						*/
/************************************************/
static int gacputenv(const void *caller,char *valeur)

{
int status;
register char *ptr;

status=-1;
if ((ptr=strchr(valeur,'='))!=(char *)0) {
  register int found;
  register int taille;
  register int tocheck;

  found=0;
  taille=0;
  if ((tocheck=(ptr-valeur))<=0)
    return status;
  status=0;
  if (environ!=(char **)0) {
    for (;environ[taille]!=(char *)0;taille++) {
      if (strncmp(valeur,environ[taille],tocheck)==0) {
        (void) gacfree(caller,environ[taille]);
	environ[taille]=valeur;
	found=1;
	break;
	}
      }
    }
  else
    environ=gaccalloc(caller,1,sizeof(char *));
  if (found==0) {
    environ=(char **)gacrealloc(caller,(void *)environ,(taille+2)*sizeof(char *));
    environ[taille]=valeur;
    environ[taille+1]=(char *)0;
    }
  }
return status;
}
/*

*/
/************************************************/
/*						*/
/*	Procedure to allocate a string within	*/
/*	a dynamic memory using a va_list	*/
/*						*/
/************************************************/
static int gacvasprintf(const void *caller,char **strp, const char *fmt,va_list args)

{
int count;
va_list argscpy;
char min[2];

count=-1;
va_copy(argscpy,args);
count=vsnprintf(min,sizeof(min),fmt,args);
*strp=gaccalloc(caller,count+1,sizeof(char));
count=vsnprintf(*strp,count+1,fmt,argscpy);
va_end(argscpy);
return count;
}
/*

*/
/************************************************/
/*						*/
/*	Procedure to initialise various GAC	*/
/*	variable.				*/
/*						*/
/************************************************/
__attribute__((constructor)) void gac_begin(void)

{
Dl_info info;
void *zone[MXTB];

(void) setcodelim();
(void) strcpy(logdir,LOGDIR);
if (dladdr(codefirst,&info)!=0) {
  (void) memset(cpath,'\000',sizeof(cpath));
  (void) strncpy(cpath,info.dli_fname,sizeof(cpath)-1);
  if (strlen(info.dli_fname)!=strlen(cpath)) {
    (void) fprintf(stderr,"Fatal! Application path <%s> is too long to be stored in curpath\n",info.dli_fname);
    (void) kill(getpid(),SIGSEGV);
    (void) exit(-1);	/*never reached		*/
  
    }
  }
/*to initialise backtrace			*/
(void) backtrace(zone,MXTB-1);
}
/*

*/
/************************************************/
/*						*/
/*	Procedure to free various malloc	*/
/*	pointer.				*/
/*						*/
/************************************************/
__attribute__((destructor)) void gac_end(void)

{
}
/*

*/
/************************************************/
/*						*/
/*	Procedure to dump and display memory	*/
/*	remaining in the garbage collector	*/
/*						*/
/************************************************/
static void dumpmem(GACMEM *gac,char *comment)

{
#define	DUMPFRM	"%s/memoryleak.%s%s%s.%05d"

FILE *fichier;
LEAKTYP *leak;

fichier=openrptfile(gac);
if (collected>0) {
  register int i;

  (void) fprintf(fichier,"%05ld memory allocation still open\n",collected);
  (void) qsort(memsort,collected,sizeof(LEAKTYP *),cmpleak);
  leak=(LEAKTYP *)0;
  for (i=0;i<collected;i++) {
    #define STRTIME	"%Y-%m-%d %H:%M:%S"
    struct tm tm;
    char strtime[40];

    if ((leak==(LEAKTYP *)0)||(sametrace(leak,memsort[i])!=0)) {
      (void) fprintf(fichier,"\n");
      leak=memsort[i];
      (void) disline(leak,cpath,fichier,gac->extension);
      }
    (void) localtime_r(&(memsort[i]->date.tv_sec),&tm);
    (void) strftime(strtime,sizeof(strtime),STRTIME,&tm);
    (void) fprintf(fichier,"\t size=%05ld age='%s.%06ld' ptr=%08lx->'%s'\n",
		     	   (long)(memsort[i]->size),
			   strtime,memsort[i]->date.tv_usec,
			   (long)(memsort[i]->ptr),
			   (char *)(memsort[i]->ptr));
    }
  if ((comment!=(char *)0)&&(strlen(comment)>0)) 
    (void) fprintf(fichier,"%s\n",comment);
  }
else 
  (void) fprintf(fichier,"No memory leak detected\n");
if (fichier!=stderr)
  (void) fclose(fichier);
}
/*

*/
/************************************************/
/*						*/
/*	Procedure to return the number of byte	*/
/*	still allocated according mallinfo.	*/
/*						*/
/************************************************/
static long getremaining(GACMEM *gac)

{
long remaining;

remaining=0;
if (gac!=(GACMEM *)0) 
  remaining=mallinfo().uordblks-gac->memstart;
return remaining;
}
/*

*/
/************************************************/
/*						*/
/*	Procedure to start collecting		*/
/*	date-memory allocation.			*/
/*						*/
/************************************************/
static int monitoring_start(GACTYP *gat,const void *caller)

{
int status;
GACMEM *gac;
int phase;
int proceed;
int hookset;

status=-1;
gac=(GACMEM *)gat;
phase=0;
proceed=true;
while (proceed==true) {
  switch (phase) {
    case 0	:	/*open ?		*/
      if (gac==(GACMEM *)0) {
	errno=0;	/*is not open		*/
	phase=999;	/*no trouble		*/
	}
      break;
    case 1	:	/*allready started	*/
      if (gac->ready==true) {
	errno=EBUSY;
	phase=999;	/*double start		*/
	}
      break;
    case 2	:	/*switch to libc alloc	*/
      (void) switchhook(&hookset,caller);
      break;
    case 3	:	/*adding GC to the list*/
      gac->ready=true;
      cgac=gac;
      break;
    case 4	:	/*switch to GC alloc	*/
      inhook=true;
      status=0;
      break;
    default	:	/*SAFE Guard		*/
      proceed=false;
      break;
    }
  phase++;
  }
return status;
}
/*

*/
/************************************************/
/*						*/
/*	Procedure to stop collecting		*/
/*	date-memory allocation.			*/
/*	Return -1 if trouble or the number of	*/
/*	memory still use by programm.		*/
/*						*/
/************************************************/
static long monitoring_stop(GACMEM *gac,long *remaining,const void *caller)

{
long status;
int hookset;
int phase;
int proceed;

status=-1;
hookset=false;
phase=0;
proceed=true;
while (proceed==true) {
  switch (phase) {
    case 0	:	/*open ?		*/
      if (gac==(GACMEM *)0) {
	status=0;
	errno=0;	/*is not open		*/
	phase=999;	/*no trouble		*/
	}
      break;
    case 1	:	/*last started ?	*/
      if (cgac!=gac) {
	errno=EFAULT;
	phase=999;	/* not last started	*/
	}
      break;
    case 2	:	/*was already stoped?	*/
      if (gac->ready==false) {
	errno=EBADF;
	phase=999;	/*stop without start	*/
	}
      break;
    case 3	:	/*switch to libc alloc	*/
      (void) switchhook(&hookset,caller);
      break;
    case 4	:	/*stoping GC		*/
      if (gac!=(GACMEM *)0) {	/*always	*/
        cgac=gac->prev;
	hookset=gac->prvhookset;
        gac->ready=false;
	}
      break;
    case 5	:	/*switching to proper GC*/
      if (cgac==(GACMEM *)0) 
	hookset=false;
      inhook=hookset;
      *remaining=getremaining(gac);
      break;
    default	:	/*SAFE Guard		*/
      proceed=false;
      break;
    }
  phase++;
  }
return status;
}
/*

*/
/************************************************/
/*						*/
/*	Procedure to extract and return current	*/
/*	version number.				*/
/*						*/
/************************************************/
const char *gac_getversion()

{
return CURTAG;
}
/*

*/
/************************************************/
/*						*/
/*	Procedure to memorise the current	*/
/*	application base, will be used to 	*/
/*	dump the memory stat.			*/
/*						*/
/************************************************/
void gac_setcurbase(int *debug,int *verbose,char *path)

{
if (debug!=(int *)0)
  dbg=debug;
if (verbose!=(int *)0)
  vbo=verbose;
if (path!=(char *)0) {
  (void) memset(cbase,'\000',sizeof(cbase));
  (void) strncpy(cbase,path,sizeof(cbase)-1);
  }
}
/*

*/
/************************************************/
/*						*/
/*	Procedure to memorise the log directory	*/
/*	to dump memory leak alarm		*/
/*						*/
/************************************************/
void gac_setlogdir(const char *path)

{
(void) memset(logdir,'\000',sizeof(logdir));
(void) strncpy(logdir,path,sizeof(logdir)-1);
}
/*

*/
/************************************************/
/*						*/
/*	Procedure to open a garbage collector	*/
/*						*/
/************************************************/
GACTYP *gac_open(int tgtdbg,const char *appname,const char *extension)

{
GACMEM *gac;
const void *caller;
int hookset;
int phase;
int proceed;

gac=(GACMEM *)0;
caller=(const void *)0;
hookset=false;
phase=0;
proceed=true;
while (proceed==true) {
  switch (phase) {
    case 0	:	/*dbg level high enough	*/
      if (tgtdbg>*dbg) {
	errno=0;	/*success		*/
	phase=999;	/*no need to open	*/
	}
      break;
    case 1	:	/*switch to libc alloc	*/
      caller=__builtin_return_address(1);
      (void) switchhook(&hookset,caller);
      break;
    case 2	:	/*making the structure	*/
      if ((gac=(GACMEM *)calloc(1,sizeof(GACMEM)))!=(GACMEM *)0) {
	gac->ready=false;
	gac->tomemo=0;
	gac->tosize=0;
	gac->prvhookset=hookset;
	if ((appname==(char *)0)&&(strlen(appname)==0))
	  appname=APPNAME;
	gac->appname=strdup(appname);
	if (extension==(char *)0)
	  extension="";
	gac->extension=strdup(extension);
	gac->tgtdbg=tgtdbg;
	}
      else {
	errno=ENOMEM;
	phase=999;	/*no need to open	*/
	}
      break;
    case 3	:	/*making the structure	*/
      (void) addinggac(&gaclst,gac);
      break;
    case 4	:	/*starting monitoring	*/
      gac->memstart=mallinfo().uordblks;
      (void) monitoring_start(gac,caller);
      break;
    default	:	/*SAFE Guard		*/
      proceed=false;
      break;
    }
  phase++;
  }
return (GACTYP *)gac;
}
/*

*/
/************************************************/
/*						*/
/*	Procedure to close a garbage collector	*/
/*	should be always 0, if less than zero	*/
/*	there is a trouble, if more than zero	*/
/*	report discrepency betweem malloc and	*/
/*	gac.					*/
/*						*/
/************************************************/
int gac_close(GACTYP *gat,long *remaining)


{
long status;
GACMEM *gac;
const void *caller;
int hookset;
int phase;
int proceed;

status=-1;
*remaining=0;
gac=(GACMEM *)gat;
caller=(const void *)0;
hookset=false;
phase=0;
proceed=true;
while (proceed==true) {
  switch (phase) {
    case 0	:	/*GC was not opened	*/
      if (gac==(GACMEM *)0) {
	status=0;
	phase=999;	/*it was an empty GC	*/
	}
      break;
    case 1	:	/*closing gac if needed	*/
      caller=__builtin_return_address(1);
      if (gac->ready==true)
	(void) monitoring_stop(gac,remaining,caller);
      break;
    case 2	:	/*closing if needed	*/
      if (cgac==gac) 
  	*remaining=getremaining(cgac);
      break;
    case 3	:	/*switch to libc alloc	*/
      if (gac->tomemo>0) {
	errno=EUCLEAN;
	phase=999;	/*all memory not free!	*/
	}
      break;
    case 4	:	/*finding gac in list	*/
      if (remgac(&gaclst,gac)<0) {
	errno=EBADF;
	phase=999;	/*trouble trouble	*/
	}
      break;
    case 5	:	/*remaining memory	*/
      (void) switchhook(&hookset,caller);
      break;
    case 6	:	/*freeing GC memory	*/
      status=0;
      hookset=gac->prvhookset;
      (void) free(gac->extension);
      (void) free(gac->appname);
      (void) free(gac);
      if (gaclst==(GACMEM *)0) 
	hookset=false;
      inhook=hookset;
      if (*remaining!=0) 
	status=1;
      break;
    default	:	/*SAFE Guard		*/
      proceed=false;
      break;
    }
  phase++;
  }
return status;
}
/*

*/
/************************************************/
/*						*/
/*	Procedure to change which Garbage	*/
/*	collector is watching about memory.	*/
/*	if the current GC collector is the last	*/
/*	openned and gac_gear(gear_up) is called,*/
/*	memory will not be supervised until next*/
/*	gac_gear(gear_down). Such case, if you	*/
/*	assigned a memory not supervised,	*/
/*	freeing this memory after the		*/
/*	'gear_down' will trigger an alarm	*/
/*	because you are requesting to free	*/
/*	a not referenced memory.		*/
/*						*/
/************************************************/
int gac_gear(GEAR updown,long *remaining)

{
int status;

status=0;
*remaining=(long)0;
if (cgac!=(GACMEM *)0) {
  static int bottom=false;
  static int top=false;

  *remaining=getremaining(cgac);
  switch (updown) {
    case gear_up	:
      if (top==false) {
	switch (bottom) {
	  case true	:
	    bottom=false;
	    inhook=true;
	    break;
	  case false	:
	    if (cgac->next!=(GACMEM *)0) {
	      cgac=cgac->next;
	      inhook=true;
	      }
	    else {
	      top=true;
	      inhook=false;
	     }
	    break;
	  }
	}
      break;
    case gear_down	:
      if (bottom==false) {
	switch (top) {
	  case true	:
	    top=false;
	    inhook=true;
	    break;
	  case false	:
            if (cgac->prev!=(GACMEM *)0) {
	      cgac=cgac->prev;
	      inhook=true;
	      }
	    else {
	      bottom=true;
	      inhook=false;
	      }
	    break;
	  }
	}
      break;
    default	:
      status=-1;
      errno=EBADF;
      break;
    }
  }
return status;
}
/*

*/
/************************************************/
/*						*/
/*	Procedure to add a string to		*/
/*	Garabage collector report.		*/
/*	Return the number of charater written.	*/
/*						*/
/************************************************/
int gac_report(int tgtdbg,char *fmt,...)

{
int written;

written=0;
if ((cgac!=(GACMEM *)0)&&(*dbg>=tgtdbg)) {
  va_list args;
  FILE *report;

  va_start(args,fmt);
  report=openrptfile(cgac);
  (void) flock(fileno(report),LOCK_EX);
  written=vfprintf(report,fmt,args);
  (void) flock(fileno(report),LOCK_UN);
  (void) closerptfile(report);
  va_end(args);
  }
return written;
}
/*

*/
/************************************************/
/*						*/
/*	Procedure to trash/free all memory	*/
/*	referenced within the garbage collector	*/
/*	Return -1 if according malloc some	*/
/*	memory is still	assigned.		*/
/*						*/
/************************************************/
int gac_trash(int tgtdbg,GACTYP *gat,long *remaining)

{
int status;
int hookset;
GACMEM *gac;
const void *caller;
int phase;
int proceed;

status=0;
hookset=false;
gac=(GACMEM *)gat;
caller=(const void *)0;
phase=0;
proceed=true;
while (proceed==true) {
  switch (phase) {
    case 0	:	/*real G-collector?	*/
      if (gac==(GACMEM *)0) 
	phase=999;	/*it was an empty GC	*/
      break;
    case 1	:	/*closing gac if needed	*/
      caller=__builtin_return_address(1);
      if (gac->ready==true)
	(void) monitoring_stop(gac,remaining,caller);
      break;
    case 2	:	/*system mode		*/
      (void) switchhook(&hookset,caller);
      break;
    case 3	:	/*extended report	*/
      if (gac->tomemo>0) {
        memsort=(LEAKTYP **)calloc(gac->tomemo,sizeof(LEAKTYP *));
  	(void) twalk(gac->memory,collectleakptr);
	}
      else
	phase++;	/*no need to free memory*/
      if (*dbg>=tgtdbg) 
	(void) dumpmem(gac,"the memory allocation(s) listed above are now free to be used again");
      break;
    case 4	:	/*removing all memory	*/
      if (collected>0){	/*always		*/
        register LEAKTYP **lptr;

        lptr=memsort+collected;	
	do {
	  collected--;
	  lptr--;
          gac->tomemo--;
          gac->tosize-=(*lptr)->size;
	  (void) free((*lptr)->ptr);
    	  (void) tdelete(*lptr,&(gac->memory),cmpptr);
          (void) free(*lptr);
	  }
	while (collected>0);
	}
      (void) free(memsort);
      break;
    case 5	:	/*GC mode (if needed)	*/
      inhook=hookset;
      (void) gac_close(gat,remaining);
      if (*remaining!=0)
	status--;	/*just report remaining	*/
      break;
    default	:	/*SAFE Guard		*/
      proceed=false;
      break;
    }
  phase++;
  }
return status;
}
/*

*/
/************************************************/
/*						*/
/*	Procedure to allocate memory space	*/
/*	and log it in the garbage collector.	*/
/*						*/
/************************************************/
void *malloc(size_t size)

{
return gacmalloc(__builtin_return_address(0),size);
}
/*

*/
/************************************************/
/*						*/
/*	Procedure to reallocate memory space	*/
/*	and log it in the garbage collector.	*/
/*						*/
/************************************************/
void *realloc(void *ptr,size_t size)

{
return gacrealloc(__builtin_return_address(0),ptr,size);
}
/*

*/
/************************************************/
/*						*/
/*	Procedure to free memory space		*/
/*	and log it out from the garbage 	*/
/*	collector.				*/
/*						*/
/************************************************/
void free(void *ptr)

{
(void) gacfree(__builtin_return_address(0),ptr);
}
/*

*/
/************************************************/
/*						*/
/*	Procedure to allocate memory properly	*/
/*	aligned onf according boundery multiple	*/
/*						*/
/************************************************/
void *memaligne(size_t alignement,size_t size)

{
return gacmemaligne(__builtin_return_address(0),alignement,size);
}
/*

*/
/************************************************/
/*						*/
/*	Procedure to get a memory area fill with*/
/*	zero.					*/
/*						*/
/************************************************/
void *calloc(size_t nmemb,size_t size)

{
return gaccalloc(__builtin_return_address(0),nmemb,size);
}
/*

*/
/************************************************/
/*						*/
/*	Procedure to install variable within the*/
/*	environnement				*/
/*						*/
/************************************************/
int putenv(char *valeur)

{
return gacputenv(__builtin_return_address(0),valeur);
}
/*

*/
/************************************************/
/*						*/
/*	Procedure to set variable within the	*/
/*	environement				*/
/*						*/
/************************************************/
int setenv(const char *name,const char *value,int overwrite)

{
register int status;
register const void *caller;
register int taille;
char *newvar;
char *ptr;

status=0;
caller=__builtin_return_address(0);
taille=strlen(name)+strlen(value)+2;
newvar=(char *)gaccalloc(caller,taille,sizeof(char));
(void) strcpy(newvar,name);
(void) strcat(newvar,"=");
(void) strcat(newvar,value);
if ((ptr=getenv(name))!=(char *)0) {
  if (overwrite!=0) 
    (void) gacunsetenv(caller,name);
  else {
    (void) gacfree(caller,newvar);
    newvar=(char *)0;
    }
  }
if (newvar!=(char *)0) {
  (void) gacputenv(caller,newvar);
  }
return status;
}
/*

*/
/************************************************/
/*						*/
/*	Procedure to remove a variable from the	*/
/*	environnement list			*/
/*						*/
/************************************************/
int unsetenv(const char *name)

{
return gacunsetenv(__builtin_return_address(0),name);
}
/*

*/
/************************************************/
/*						*/
/*	Procedure to allocate a string within	*/
/*	a dynamic memory using a va_list	*/
/*						*/
/************************************************/
int vasprintf(char **strp, const char *fmt,va_list args)

{
return gacvasprintf(__builtin_return_address(0),strp,fmt,args);
}
/*

*/
/************************************************/
/*						*/
/*	Procedure to allocate a string within	*/
/*	a dynamic memory			*/
/*						*/
/************************************************/
int asprintf(char **strp, const char *fmt,...)

{
int count;
va_list args;

count=-1;
va_start(args,fmt);
count=gacvasprintf(__builtin_return_address(0),strp,fmt,args);
va_end(args);
return count;
}
/*

*/
/************************************************/
/*						*/
/*	Procedure to allocate memory to		*/
/*	duplicate a string.			*/
/*						*/
/************************************************/
char *strdup(const char *str)

{
register int taille;
char *ptr;

if (str==(char *)0) {
  /*NULL PTR to duplicate?! crashing HARD	*/
  (void) kill(getpid(),SIGSEGV);
  }
taille=strlen(str)+1;
ptr=gaccalloc(__builtin_return_address(0),taille,sizeof(char));
(void) memcpy(ptr,str,taille);
return ptr;
}
/*

*/
/************************************************/
/*						*/
/*	Procedure to allocate memory to		*/
/*	duplicate a string but only for a max	*/
/*	number of character.			*/
/*						*/
/************************************************/
char *strndup(const char *str,size_t n)

{
char *ptr;
register int taille;

if (str==(char *)0) {
  /*NULL PTR to duplicate?! crashing HARD	*/
  (void) kill(getpid(),SIGSEGV);
  }
taille=strlen(str);
if (taille>n)
  taille=n;
ptr=gaccalloc(__builtin_return_address(0),taille+1,sizeof(char));
(void) memcpy(ptr,str,taille);
return ptr;
}
